#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define WARP_SIZE 32
#define BLOCK_SIZE 256
#define TILE_SIZE 16

// Device function for computing convolution at a single point
__device__ __forceinline__ float compute_conv_point(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    int ih, int iw, int H, int W,
    int C, int kh, int kw) {
    if (ih >= 0 && ih < H && iw >= 0 && iw < W) {
        return input[ih * W + iw] * weight[kh * 3 + kw];
    }
    return 0.0f;
}

// Device function for computing bias and ReLU
__device__ __forceinline__ float bias_relu(float x, float bias_val) {
    return max(x + bias_val, 0.0f);
}

// Device function for computing convolution sum
__device__ __forceinline__ float compute_conv_sum(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    int h, int w, int H, int W,
    int C, int P) {
    float sum = 0.0f;
    
    #pragma unroll
    for (int c = 0; c < C; ++c) {
        const float* input_c = input + c * H * W;
        const float* weight_c = weight + c * 9;  // 3x3 kernel
        
        #pragma unroll
        for (int kh = 0; kh < 3; ++kh) {
            #pragma unroll
            for (int kw = 0; kw < 3; ++kw) {
                int ih = h - P + kh;
                int iw = w - P + kw;
                sum += compute_conv_point(input_c, weight_c, ih, iw, H, W, C, kh, kw);
            }
        }
    }
    return sum;
}

// Main convolution kernel
template<int CHANNELS>
__global__ void conv2d_modular_kernel(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int N, int H, int W, int K, int P) {
    
    extern __shared__ float shared_mem[];
    
    const int tid = threadIdx.y * blockDim.x + threadIdx.x;
    const int n = blockIdx.z;
    const int k = blockIdx.y;
    const int h = blockIdx.x * blockDim.y + threadIdx.y;
    const int w = threadIdx.x;
    
    if (k >= K || h >= H || w >= W) return;
    
    // Compute convolution for this position
    const float* input_n = input + n * CHANNELS * H * W;
    const float* weight_k = weight + k * CHANNELS * 9;
    float sum = compute_conv_sum(input_n, weight_k, h, w, H, W, CHANNELS, P);
    
    // Apply bias and ReLU
    sum = bias_relu(sum, bias[k]);
    
    // Write result
    output[n * K * H * W + k * H * W + h * W + w] = sum;
}

torch::Tensor vgg16_forward_cuda(
    torch::Tensor x,
    std::vector<torch::Tensor> conv_weights,
    std::vector<torch::Tensor> conv_biases,
    std::vector<torch::Tensor> fc_weights,
    std::vector<torch::Tensor> fc_biases,
    bool is_training
) {
    const int N = x.size(0);
    const int C = x.size(1);
    const int H = x.size(2);
    const int W = x.size(3);
    const int K = conv_weights[0].size(0);
    const int P = 1;  // padding

    auto current = torch::empty({N, K, H, W}, x.options());

    dim3 block(TILE_SIZE, TILE_SIZE);
    dim3 grid((W + TILE_SIZE - 1) / TILE_SIZE,
              (K + TILE_SIZE - 1) / TILE_SIZE,
              N);

    // Launch specialized kernel for first conv layer
    conv2d_modular_kernel<3><<<grid, block, TILE_SIZE * TILE_SIZE * sizeof(float)>>>(
        x.data_ptr<float>(),
        conv_weights[0].data_ptr<float>(),
        conv_biases[0].data_ptr<float>(),
        current.data_ptr<float>(),
        N, H, W, K, P
    );

    current = torch::relu(current);
    
    // Rest of the network using optimized cudnn calls
    for (int i = 1; i < 13; ++i) {
        current = torch::conv2d(current, conv_weights[i], conv_biases[i], 
                              /*stride=*/1, /*padding=*/1);
        current = torch::relu(current);
        if (i == 1 || i == 3 || i == 6 || i == 9 || i == 12) {
            current = torch::max_pool2d(current, /*kernel_size=*/2, /*stride=*/2);
        }
    }

    // Fully connected layers
    current = current.flatten(1);
    for (int i = 0; i < 3; ++i) {
        current = torch::linear(current, fc_weights[i], fc_biases[i]);
        if (i < 2) {
            current = torch::relu(current);
            if (is_training) {
                current = torch::dropout(current, /*p=*/0.0, /*train=*/true);
            }
        }
    }

    return current;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &vgg16_forward_cuda, "VGG16 forward (CUDA)");
}